package ru.dbotthepony.mc.otm;

import net.minecraft.ChatFormatting;
import net.minecraft.client.gui.screens.MenuScreens;
import net.minecraft.client.renderer.blockentity.BlockEntityRenderers;
import net.minecraft.network.chat.Component;
import net.minecraft.network.chat.TextComponent;
import net.minecraft.network.chat.TranslatableComponent;
import net.minecraft.resources.ResourceLocation;
import net.minecraft.stats.StatFormatter;
import net.minecraft.tags.BlockTags;
import net.minecraft.tags.ItemTags;
import net.minecraft.util.valueproviders.UniformInt;
import net.minecraft.world.damagesource.DamageSource;
import net.minecraft.world.entity.EquipmentSlot;
import net.minecraft.world.food.FoodProperties;
import net.minecraft.world.inventory.MenuType;
import net.minecraft.world.item.*;
import net.minecraft.world.item.crafting.Ingredient;
import net.minecraft.world.item.crafting.Recipe;
import net.minecraft.world.item.crafting.RecipeSerializer;
import net.minecraft.world.item.crafting.RecipeType;
import net.minecraft.world.level.Level;
import net.minecraft.world.level.block.Block;
import net.minecraft.world.level.block.OreBlock;
import net.minecraft.world.level.block.SoundType;
import net.minecraft.world.level.block.entity.BlockEntityType;
import net.minecraft.world.level.block.state.BlockBehaviour;
import net.minecraft.world.level.material.Material;
import net.minecraft.world.level.material.MaterialColor;
import net.minecraftforge.common.ForgeTier;
import net.minecraftforge.event.RegistryEvent;
import net.minecraftforge.eventbus.api.SubscribeEvent;
import net.minecraftforge.fml.event.lifecycle.FMLClientSetupEvent;
import net.minecraftforge.fml.event.lifecycle.FMLCommonSetupEvent;
import net.minecraftforge.registries.ForgeRegistry;
import net.minecraftforge.registries.RegistryBuilder;
import org.jetbrains.annotations.Nullable;
import ru.dbotthepony.mc.otm.android.AndroidFeature;
import ru.dbotthepony.mc.otm.android.AndroidResearchBuilder;
import ru.dbotthepony.mc.otm.android.AndroidResearchType;
import ru.dbotthepony.mc.otm.android.feature.AndroidExtendedReach;
import ru.dbotthepony.mc.otm.android.feature.AndroidLimbOverclocking;
import ru.dbotthepony.mc.otm.android.feature.AndroidNanobotsArmor;
import ru.dbotthepony.mc.otm.android.feature.AndroidNanobotsRegeneration;
import ru.dbotthepony.mc.otm.block.*;
import ru.dbotthepony.mc.otm.block.entity.*;
import ru.dbotthepony.mc.otm.android.AndroidFeatureType;
import ru.dbotthepony.mc.otm.block.entity.blackhole.BlockEntityBlackHole;
import ru.dbotthepony.mc.otm.block.entity.blackhole.BlockEntityExplosionDebugger;
import ru.dbotthepony.mc.otm.block.entity.blackhole.BlockEntitySphereDebugger;
import ru.dbotthepony.mc.otm.client.render.BlackHoleRenderer;
import ru.dbotthepony.mc.otm.client.render.GravitationStabilizerRenderer;
import ru.dbotthepony.mc.otm.client.render.SkinElement;
import ru.dbotthepony.mc.otm.core.Fraction;
import ru.dbotthepony.mc.otm.item.*;
import ru.dbotthepony.mc.otm.menu.*;
import ru.dbotthepony.mc.otm.client.screen.*;
import ru.dbotthepony.mc.otm.recipe.PlatePressRecipe;
import ru.dbotthepony.mc.otm.recipe.PlatePressRecipeFactory;

import net.minecraftforge.common.Tags.IOptionalNamedTag;

import javax.annotation.Nonnull;
import java.util.List;
import java.util.Set;

public class Registry {
	public static final DamageSource DAMAGE_BECOME_ANDROID = new DamageSource("otm_become_android");
	public static final DamageSource DAMAGE_BECOME_HUMANE = new DamageSource("otm_become_humane");
	public static final DamageSource DAMAGE_EVENT_HORIZON = new DamageSource("otm_event_horizon");
	public static final DamageSource DAMAGE_HAWKING_RADIATION = new DamageSource("otm_hawking_radiation");

	static {
		DAMAGE_BECOME_ANDROID.bypassArmor().bypassInvul().bypassMagic();
		DAMAGE_BECOME_HUMANE.bypassArmor().bypassInvul().bypassMagic();
		DAMAGE_EVENT_HORIZON.bypassMagic().bypassArmor();
		// DAMAGE_HAWKING_RADIATION.bypassMagic().bypassArmor();
	}

	public static final ForgeRegistry<AndroidFeatureType<?>> ANDROID_FEATURES;
	public static final ForgeRegistry<AndroidResearchType<?>> ANDROID_RESEARCH;

	static {
		var builder = new RegistryBuilder<AndroidFeatureType<?>>();
		builder.setName(new ResourceLocation(OverdriveThatMatters.MOD_ID, "android_features"));

		// make it shut up
		builder.setType(c(AndroidFeatureType.class));
		ANDROID_FEATURES = (ForgeRegistry<AndroidFeatureType<?>>) builder.create();

		var builder2 = new RegistryBuilder<AndroidResearchType<?>>();
		builder2.setName(new ResourceLocation(OverdriveThatMatters.MOD_ID, "android_research"));

		// make it shut up
		builder2.setType(c(AndroidResearchType.class));
		ANDROID_RESEARCH = (ForgeRegistry<AndroidResearchType<?>>) builder2.create();
	}

	public static void dummy() {}

	@SuppressWarnings("unchecked")
	private static <T> Class<T> c(Class<?> cls) { return (Class<T>) cls; }

	record CrateProps(MaterialColor color, ResourceLocation name) {
		public CrateProps(MaterialColor color, String name) {
			this(color, new ResourceLocation(OverdriveThatMatters.MOD_ID, "crate_" + name));
		}
	}

	public static final CrateProps[] CRATES = new CrateProps[] {
		new CrateProps(MaterialColor.COLOR_RED, "red"),
		new CrateProps(MaterialColor.COLOR_BLUE, "blue"),
		new CrateProps(MaterialColor.COLOR_YELLOW, "yellow"),
		new CrateProps(MaterialColor.COLOR_GREEN, "green"),
		new CrateProps(MaterialColor.COLOR_BLACK, "black"),
		new CrateProps(MaterialColor.COLOR_PINK, "pink"),
		new CrateProps(MaterialColor.COLOR_PURPLE, "purple"),
	};

	public static class Names {
		private static ResourceLocation loc(String path) {
			return new ResourceLocation(OverdriveThatMatters.MOD_ID, path);
		}

		// blocks
		public static final ResourceLocation ANDROID_STATION = loc("android_station"); // без рецепта
		public static final ResourceLocation BATTERY_BANK = loc("battery_bank"); // нужен рецепт
		public static final ResourceLocation MATTER_DECOMPOSER = loc("matter_decomposer"); // есть рецепт
		public static final ResourceLocation MATTER_CAPACITOR_BANK = loc("matter_capacitor_bank"); // есть рецепт
		public static final ResourceLocation MATTER_CABLE = loc("matter_cable"); // есть рецепт
		public static final ResourceLocation PATTERN_STORAGE = loc("pattern_storage"); // есть рецепт
		public static final ResourceLocation MATTER_SCANNER = loc("matter_scanner"); // есть рецепт
		public static final ResourceLocation MATTER_PANEL = loc("matter_panel"); // есть рецепт
		public static final ResourceLocation MATTER_REPLICATOR = loc("matter_replicator"); // есть рецепт
		public static final ResourceLocation MATTER_BOTTLER = loc("matter_bottler"); // есть рецепт
		public static final ResourceLocation DRIVE_VIEWER = loc("drive_viewer"); // есть рецепт
		public static final ResourceLocation DRIVE_RACK = loc("drive_rack"); // нужен рецепт (после улучшений)
		public static final ResourceLocation ITEM_MONITOR = loc("item_monitor"); // нужен рецепт (после улучшений)
		public static final ResourceLocation ENERGY_COUNTER = loc("energy_counter"); // есть рецепт
		public static final ResourceLocation CHEMICAL_GENERATOR = loc("chemical_generator"); // есть рецепт
		public static final ResourceLocation PLATE_PRESS = loc("plate_press"); // есть рецепт


		public static final ResourceLocation DEBUG_EXPLOSION_SMALL = loc("debug_explosion_small");
		public static final ResourceLocation DEBUG_SPHERE_POINTS = loc("debug_sphere_points");

		public static final ResourceLocation BLACK_HOLE = loc("black_hole");
		public static final ResourceLocation GRAVITATION_STABILIZER = loc("gravitation_stabilizer"); // нужен рецепт
		public static final ResourceLocation CARGO_CRATE = loc("cargo_crate"); // нужен рецепт?

		// building blocks
		public static final ResourceLocation TRITANIUM_BLOCK = loc("tritanium_block");
		public static final ResourceLocation TRITANIUM_STRIPED_BLOCK = loc("tritanium_striped_block");

		public static final ResourceLocation CARBON_FIBRE_BLOCK = loc("carbon_fibre_block");

		// capabilities
		public static final ResourceLocation ANDROID_CAPABILITY = loc("android_capability");

		// items
		public static final ResourceLocation GRAVITATIONAL_DISRUPTOR = loc("gravitational_disruptor");

		public static final ResourceLocation PILL_ANDROID = loc("pill_android");
		public static final ResourceLocation PILL_HUMANE = loc("pill_humane");
		public static final ResourceLocation PILL_OBLIVION = loc("pill_oblivion");
		public static final ResourceLocation PILL_HEAL = loc("pill_heal");

		public static final ResourceLocation BATTERY_CRUDE = loc("battery_crude");
		public static final ResourceLocation BATTERY_BASIC = loc("battery_basic");
		public static final ResourceLocation BATTERY_NORMAL = loc("battery_normal");
		public static final ResourceLocation BATTERY_DENSE = loc("battery_dense");
		public static final ResourceLocation BATTERY_CAPACITOR = loc("battery_capacitor");
		public static final ResourceLocation BATTERY_CREATIVE = loc("battery_creative");

		public static final ResourceLocation MATTER_CAPACITOR_PARTS = loc("matter_capacitor_parts");
		public static final ResourceLocation MATTER_CAPACITOR_BASIC = loc("matter_capacitor_basic");
		public static final ResourceLocation MATTER_CAPACITOR_NORMAL = loc("matter_capacitor_normal");
		public static final ResourceLocation MATTER_CAPACITOR_DENSE = loc("matter_capacitor_dense");
		public static final ResourceLocation MATTER_CAPACITOR_CREATIVE = loc("matter_capacitor_creative");

		public static final ResourceLocation PATTERN_DRIVE_NORMAL = loc("pattern_drive_normal");
		public static final ResourceLocation PATTERN_DRIVE_CREATIVE = loc("pattern_drive_creative");

		public static final ResourceLocation NUTRIENT_PASTE = loc("nutrient_paste");

		public static final ResourceLocation PORTABLE_CONDENSATION_DRIVE = loc("portable_condensation_drive");
		public static final ResourceLocation PORTABLE_DENSE_CONDENSATION_DRIVE = loc("portable_dense_condensation_drive");

		public static final ResourceLocation BLACK_HOLE_SCANNER = loc("black_hole_scanner");
		public static final ResourceLocation GRAVITATION_FIELD_LIMITER = loc("gravitation_field_limiter");
		public static final ResourceLocation GRAVITATION_FIELD_SENSOR = loc("graivtation_field_sensor");
		public static final ResourceLocation PORTABLE_GRAVITATION_STABILIZER = loc("portable_gravitation_stabilizer");

		// armor
		public static final ResourceLocation TRITANIUM_HELMET = loc("tritanium_helmet");
		public static final ResourceLocation TRITANIUM_CHESTPLATE = loc("tritanium_chestplate");
		public static final ResourceLocation TRITANIUM_PANTS = loc("tritanium_pants");
		public static final ResourceLocation TRITANIUM_BOOTS = loc("tritanium_boots");

		// tools
		public static final ResourceLocation TRITANIUM_SWORD = loc("tritanium_sword");
		public static final ResourceLocation TRITANIUM_AXE = loc("tritanium_axe");
		public static final ResourceLocation TRITANIUM_PICKAXE = loc("tritanium_pickaxe");
		public static final ResourceLocation TRITANIUM_SHOVEL = loc("tritanium_shovel");
		public static final ResourceLocation TRITANIUM_HOE = loc("tritanium_hoe");

		// items: crafting components
		public static final ResourceLocation TRITANIUM_INGOT = loc("tritanium_ingot");

		public static final ResourceLocation MATTER_IO_PORT = loc("matter_io_port");
		public static final ResourceLocation MATTER_TRANSFORM_MATRIX = loc("matter_transform_matrix");

		public static final ResourceLocation ENERGY_BUS = loc("energy_bus");
		public static final ResourceLocation ELECTRIC_PARTS = loc("electric_parts");

		public static final ResourceLocation MACHINE_FRAME = loc("machine_frame");
		public static final ResourceLocation TRITANIUM_PLATE = loc("tritanium_plate");
		public static final ResourceLocation IRON_PLATE = loc("iron_plate");
		public static final ResourceLocation COPPER_WIRING = loc("copper_wiring");
		public static final ResourceLocation GOLD_WIRING = loc("gold_wiring");

		public static final ResourceLocation PORTABLE_CONDENSATION_DRIVE_CASING = loc("portable_condensation_drive_casing");
		public static final ResourceLocation PORTABLE_DENSE_CONDENSATION_DRIVE_CASING = loc("portable_dense_condensation_drive_casing");

		public static final ResourceLocation CIRCUIT_PLATING = loc("circuit_plating");
		public static final ResourceLocation BASIC_CONTROL_CIRCUIT = loc("basic_control_circuit");
		public static final ResourceLocation ADVANCED_CONTROL_CIRCUIT = loc("advanced_control_circuit");

		public static final ResourceLocation TRITANIUM_ORE = loc("tritanium_ore");
		public static final ResourceLocation DEEPSLATE_TRITANIUM_ORE = loc("deepslate_tritanium_ore");
		public static final ResourceLocation TRITANIUM_RAW_BLOCK = loc("tritanium_raw_block");
		public static final ResourceLocation TRITANIUM_ORE_CLUMP = loc("tritanium_ore_clump");

		// android features and research
		public static final ResourceLocation AIR_BAGS = loc("air_bags");

		public static final ResourceLocation LIMB_OVERCLOCKING = loc("limb_overclocking");
		public static final ResourceLocation LIMB_OVERCLOCKING_1 = loc("limb_overclocking_1");
		public static final ResourceLocation LIMB_OVERCLOCKING_2 = loc("limb_overclocking_2");
		public static final ResourceLocation LIMB_OVERCLOCKING_3 = loc("limb_overclocking_3");
		public static final ResourceLocation LIMB_OVERCLOCKING_4 = loc("limb_overclocking_4");

		public static final ResourceLocation[] LIMB_OVERCLOCKING_LIST = new ResourceLocation[] { LIMB_OVERCLOCKING_1, LIMB_OVERCLOCKING_2, LIMB_OVERCLOCKING_3, LIMB_OVERCLOCKING_4 };

		public static final ResourceLocation NANOBOTS = loc("nanobots");
		public static final ResourceLocation NANOBOTS_REGENERATION = loc("nanobots_regeneration");
		public static final ResourceLocation NANOBOTS_REGENERATION_1 = loc("nanobots_regeneration_1");
		public static final ResourceLocation NANOBOTS_REGENERATION_2 = loc("nanobots_regeneration_2");
		public static final ResourceLocation NANOBOTS_REGENERATION_3 = loc("nanobots_regeneration_3");
		public static final ResourceLocation NANOBOTS_REGENERATION_4 = loc("nanobots_regeneration_4");

		public static final ResourceLocation[] NANOBOTS_REGENERATION_LIST = new ResourceLocation[] { NANOBOTS_REGENERATION_1, NANOBOTS_REGENERATION_2, NANOBOTS_REGENERATION_3, NANOBOTS_REGENERATION_4 };

		public static final ResourceLocation NANOBOTS_ARMOR = loc("nanobots_armor");

		public static final ResourceLocation NANOBOTS_ARMOR_STRENGTH_1 = loc("nanobots_armor_strength_1");
		public static final ResourceLocation NANOBOTS_ARMOR_STRENGTH_2 = loc("nanobots_armor_strength_2");
		public static final ResourceLocation NANOBOTS_ARMOR_STRENGTH_3 = loc("nanobots_armor_strength_3");
		public static final ResourceLocation[] NANOBOTS_ARMOR_STRENGTH_LIST = new ResourceLocation[] {
			NANOBOTS_ARMOR_STRENGTH_1,
			NANOBOTS_ARMOR_STRENGTH_2,
			NANOBOTS_ARMOR_STRENGTH_3,
		};

		public static final ResourceLocation NANOBOTS_ARMOR_SPEED_1 = loc("nanobots_armor_speed_1");
		public static final ResourceLocation NANOBOTS_ARMOR_SPEED_2 = loc("nanobots_armor_speed_2");
		public static final ResourceLocation NANOBOTS_ARMOR_SPEED_3 = loc("nanobots_armor_speed_3");
		public static final ResourceLocation[] NANOBOTS_ARMOR_SPEED_LIST = new ResourceLocation[] {
			NANOBOTS_ARMOR_SPEED_1,
			NANOBOTS_ARMOR_SPEED_2,
			NANOBOTS_ARMOR_SPEED_3,
		};

		public static final ResourceLocation HYDRAULICS_OVERLOAD = loc("hydraulics_overload");
		public static final ResourceLocation HYDRAULICS_OVERLOAD_1 = loc("hydraulics_overload_1");
		public static final ResourceLocation HYDRAULICS_OVERLOAD_2 = loc("hydraulics_overload_2");
		public static final ResourceLocation HYDRAULICS_OVERLOAD_3 = loc("hydraulics_overload_3");

		public static final ResourceLocation EXTENDED_REACH = loc("extended_reach");

		// stats
		public static final ResourceLocation DAMAGE_ABSORBED = loc("damage_absorbed");
		public static final ResourceLocation HEALTH_REGENERATED = loc("health_regenerated");
		public static final ResourceLocation POWER_CONSUMED = loc("power_consumed");
	}

	public static class Blocks {
		public static final BlockAndroidStation ANDROID_STATION = new BlockAndroidStation();
		public static final BlockBatteryBank BATTERY_BANK = new BlockBatteryBank();
		public static final BlockMatterDecomposer MATTER_DECOMPOSER = new BlockMatterDecomposer();
		public static final BlockMatterCapacitorBank MATTER_CAPACITOR_BANK = new BlockMatterCapacitorBank();
		public static final BlockMatterCable MATTER_CABLE = new BlockMatterCable();
		public static final BlockPatternStorage PATTERN_STORAGE = new BlockPatternStorage();
		public static final BlockMatterScanner MATTER_SCANNER = new BlockMatterScanner();
		public static final BlockMatterPanel MATTER_PANEL = new BlockMatterPanel();
		public static final BlockMatterReplicator MATTER_REPLICATOR = new BlockMatterReplicator();
		public static final BlockMatterBottler MATTER_BOTTLER = new BlockMatterBottler();
		public static final BlockDriveViewer DRIVE_VIEWER = new BlockDriveViewer();
		public static final BlockCargoCrate CARGO_CRATE = new BlockCargoCrate();
		public static final BlockDriveRack DRIVE_RACK = new BlockDriveRack();
		public static final BlockItemMonitor ITEM_MONITOR = new BlockItemMonitor();
		public static final BlockEnergyCounter ENERGY_COUNTER = new BlockEnergyCounter();
		public static final BlockChemicalGenerator CHEMICAL_GENERATOR = new BlockChemicalGenerator();
		public static final BlockPlatePress PLATE_PRESS = new BlockPlatePress();

		public static final BlockExplosionDebugger DEBUG_EXPLOSION_SMALL = new BlockExplosionDebugger();
		public static final BlockSphereDebugger DEBUG_SPHERE_POINTS = new BlockSphereDebugger();

		public static final BlockBlackHole BLACK_HOLE = new BlockBlackHole();
		public static final BlockGravitationStabilizer GRAVITATION_STABILIZER = new BlockGravitationStabilizer();

		public static final Block TRITANIUM_ORE = new OreBlock(
			BlockBehaviour.Properties.of(Material.STONE)
			.strength(3.25F, 6.0F)
			.requiresCorrectToolForDrops(),
			UniformInt.of(0, 3)
		);

		public static final Block DEEPSLATE_TRITANIUM_ORE = new OreBlock(
			BlockBehaviour.Properties.of(Material.STONE)
			.strength(4.75F, 6.5F)
			.requiresCorrectToolForDrops(),
			UniformInt.of(0, 3)
		);

		public static final Block TRITANIUM_RAW_BLOCK = new Block(
			BlockBehaviour.Properties.of(Material.STONE)
			.strength(8.0F, 10F)
			.requiresCorrectToolForDrops()
		);

		public static final Block[] CRATES = new Block[Registry.CRATES.length];

		public static final Block TRITANIUM_BLOCK = new Block(
			BlockBehaviour.Properties.of(Material.METAL, MaterialColor.COLOR_LIGHT_BLUE)
				.sound(SoundType.BASALT)
				.requiresCorrectToolForDrops()
				.explosionResistance(80)
				.strength(4)
		);

		public static final Block TRITANIUM_STRIPED_BLOCK = new Block(
			BlockBehaviour.Properties.of(Material.METAL, MaterialColor.COLOR_LIGHT_BLUE)
				.sound(SoundType.BASALT)
				.requiresCorrectToolForDrops()
				.explosionResistance(80)
				.strength(4)
		);

		public static final Block CARBON_FIBRE_BLOCK = new Block(
			BlockBehaviour.Properties.of(Material.METAL, MaterialColor.COLOR_LIGHT_BLUE)
				.sound(SoundType.BASALT)
				.requiresCorrectToolForDrops()
				.explosionResistance(40)
				.strength(3)
		);

		static {
			for (int i = 0; i < Registry.CRATES.length; i++) {
				CRATES[i] = new Block(BlockBehaviour.Properties.of(Material.METAL, Registry.CRATES[i].color)
					.sound(SoundType.METAL)
					.requiresCorrectToolForDrops()
					.strength(5.0F, 6.0F));

				CRATES[i].setRegistryName(Registry.CRATES[i].name);
			}

			TRITANIUM_ORE.setRegistryName(Names.TRITANIUM_ORE);
			DEEPSLATE_TRITANIUM_ORE.setRegistryName(Names.DEEPSLATE_TRITANIUM_ORE);
			TRITANIUM_RAW_BLOCK.setRegistryName(Names.TRITANIUM_RAW_BLOCK);
			ANDROID_STATION.setRegistryName(Names.ANDROID_STATION);
			BATTERY_BANK.setRegistryName(Names.BATTERY_BANK);
			MATTER_DECOMPOSER.setRegistryName(Names.MATTER_DECOMPOSER);
			MATTER_CAPACITOR_BANK.setRegistryName(Names.MATTER_CAPACITOR_BANK);
			MATTER_CABLE.setRegistryName(Names.MATTER_CABLE);
			PATTERN_STORAGE.setRegistryName(Names.PATTERN_STORAGE);
			MATTER_SCANNER.setRegistryName(Names.MATTER_SCANNER);
			MATTER_PANEL.setRegistryName(Names.MATTER_PANEL);
			MATTER_REPLICATOR.setRegistryName(Names.MATTER_REPLICATOR);
			MATTER_BOTTLER.setRegistryName(Names.MATTER_BOTTLER);
			DRIVE_VIEWER.setRegistryName(Names.DRIVE_VIEWER);
			BLACK_HOLE.setRegistryName(Names.BLACK_HOLE);
			CARGO_CRATE.setRegistryName(Names.CARGO_CRATE);
			DRIVE_RACK.setRegistryName(Names.DRIVE_RACK);
			ITEM_MONITOR.setRegistryName(Names.ITEM_MONITOR);
			ENERGY_COUNTER.setRegistryName(Names.ENERGY_COUNTER);
			CHEMICAL_GENERATOR.setRegistryName(Names.CHEMICAL_GENERATOR);
			PLATE_PRESS.setRegistryName(Names.PLATE_PRESS);

			GRAVITATION_STABILIZER.setRegistryName(Names.GRAVITATION_STABILIZER);

			DEBUG_EXPLOSION_SMALL.setRegistryName(Names.DEBUG_EXPLOSION_SMALL);
			DEBUG_SPHERE_POINTS.setRegistryName(Names.DEBUG_SPHERE_POINTS);

			TRITANIUM_BLOCK.setRegistryName(Names.TRITANIUM_BLOCK);
			TRITANIUM_STRIPED_BLOCK.setRegistryName(Names.TRITANIUM_STRIPED_BLOCK);
			CARBON_FIBRE_BLOCK.setRegistryName(Names.CARBON_FIBRE_BLOCK);
		}

		@SubscribeEvent
		@SuppressWarnings("unused")
		public static void register(final RegistryEvent.Register<Block> event) {
			event.getRegistry().register(ANDROID_STATION);
			event.getRegistry().register(BATTERY_BANK);
			event.getRegistry().register(MATTER_DECOMPOSER);
			event.getRegistry().register(MATTER_CAPACITOR_BANK);
			event.getRegistry().register(MATTER_CABLE);
			event.getRegistry().register(PATTERN_STORAGE);
			event.getRegistry().register(MATTER_SCANNER);
			event.getRegistry().register(MATTER_PANEL);
			event.getRegistry().register(MATTER_REPLICATOR);
			event.getRegistry().register(MATTER_BOTTLER);
			event.getRegistry().register(DRIVE_VIEWER);
			event.getRegistry().register(TRITANIUM_BLOCK);
			event.getRegistry().register(TRITANIUM_STRIPED_BLOCK);
			event.getRegistry().register(CARBON_FIBRE_BLOCK);
			event.getRegistry().register(BLACK_HOLE);
			event.getRegistry().register(CARGO_CRATE);
			event.getRegistry().register(TRITANIUM_ORE);
			event.getRegistry().register(DEEPSLATE_TRITANIUM_ORE);
			event.getRegistry().register(TRITANIUM_RAW_BLOCK);
			event.getRegistry().register(DRIVE_RACK);
			event.getRegistry().register(ITEM_MONITOR);
			event.getRegistry().register(ENERGY_COUNTER);
			event.getRegistry().register(CHEMICAL_GENERATOR);
			event.getRegistry().register(PLATE_PRESS);
			event.getRegistry().register(GRAVITATION_STABILIZER);

			event.getRegistry().register(DEBUG_EXPLOSION_SMALL);
			event.getRegistry().register(DEBUG_SPHERE_POINTS);

			for (var crate : CRATES) {
				event.getRegistry().register(crate);
			}

			// OverdriveThatMatters.LOGGER.info("Registered blocks");
		}
	}

	public static class Items {
		public static final BlockItem ANDROID_STATION = new BlockItem(Blocks.ANDROID_STATION, new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));
		public static final BlockItem BATTERY_BANK = new BlockItem(Blocks.BATTERY_BANK, new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));
		public static final BlockItem MATTER_DECOMPOSER = new BlockItem(Blocks.MATTER_DECOMPOSER, new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));
		public static final BlockItem MATTER_CAPACITOR_BANK = new BlockItem(Blocks.MATTER_CAPACITOR_BANK, new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));
		public static final BlockItem MATTER_CABLE = new BlockItem(Blocks.MATTER_CABLE, new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));
		public static final BlockItem PATTERN_STORAGE = new BlockItem(Blocks.PATTERN_STORAGE, new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));
		public static final BlockItem MATTER_SCANNER = new BlockItem(Blocks.MATTER_SCANNER, new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));
		public static final BlockItem MATTER_PANEL = new BlockItem(Blocks.MATTER_PANEL, new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));
		public static final BlockItem MATTER_REPLICATOR = new BlockItem(Blocks.MATTER_REPLICATOR, new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));
		public static final BlockItem MATTER_BOTTLER = new BlockItem(Blocks.MATTER_BOTTLER, new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));
		public static final BlockItem DRIVE_VIEWER = new BlockItem(Blocks.DRIVE_VIEWER, new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));
		public static final BlockItem CARGO_CRATE = new BlockItem(Blocks.CARGO_CRATE, new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));
		public static final BlockItem TRITANIUM_ORE = new BlockItem(Blocks.TRITANIUM_ORE, new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));
		public static final BlockItem DEEPSLATE_TRITANIUM_ORE = new BlockItem(Blocks.DEEPSLATE_TRITANIUM_ORE, new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));
		public static final BlockItem TRITANIUM_RAW_BLOCK = new BlockItem(Blocks.TRITANIUM_RAW_BLOCK, new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));
		public static final BlockItem DRIVE_RACK = new BlockItem(Blocks.DRIVE_RACK, new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));
		public static final BlockItem ITEM_MONITOR = new BlockItem(Blocks.ITEM_MONITOR, new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));
		public static final BlockItem ENERGY_COUNTER = new BlockItem(Blocks.ENERGY_COUNTER, new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));
		public static final BlockItem CHEMICAL_GENERATOR = new BlockItem(Blocks.CHEMICAL_GENERATOR, new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));
		public static final BlockItem PLATE_PRESS = new BlockItem(Blocks.PLATE_PRESS, new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));
		public static final BlockItem GRAVITATION_STABILIZER = new BlockItem(Blocks.GRAVITATION_STABILIZER, new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB)) {
			@Override
			public void appendHoverText(@Nonnull ItemStack p_40572_, @Nullable Level p_40573_, @Nonnull List<Component> p_40574_, @Nonnull TooltipFlag p_40575_) {
				super.appendHoverText(p_40572_, p_40573_, p_40574_, p_40575_);

				p_40574_.add(new TranslatableComponent("block.overdrive_that_matters.gravitation_stabilizer.desc").withStyle(ChatFormatting.GRAY));
				p_40574_.add(new TranslatableComponent("block.overdrive_that_matters.gravitation_stabilizer.desc2").withStyle(ChatFormatting.DARK_GRAY));
				p_40574_.add(new TranslatableComponent("block.overdrive_that_matters.gravitation_stabilizer.desc3").withStyle(ChatFormatting.DARK_GRAY));
				p_40574_.add(new TranslatableComponent("block.overdrive_that_matters.gravitation_stabilizer.desc4").withStyle(ChatFormatting.DARK_GRAY));
			}
		};

		public static final BlockItem DEBUG_EXPLOSION_SMALL = new BlockItem(Blocks.DEBUG_EXPLOSION_SMALL, new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));
		public static final BlockItem DEBUG_SPHERE_POINTS = new BlockItem(Blocks.DEBUG_SPHERE_POINTS, new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));

		public static final Item TRITANIUM_ORE_CLUMP = new Item(new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));
		public static final Item TRITANIUM_INGOT = new Item(new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));

		public static final ForgeTier TRITANIUM_COMPONENT;

		static {
			final var component = Ingredient.of(TRITANIUM_INGOT);

			TRITANIUM_COMPONENT = new ForgeTier(
				Tiers.IRON.getLevel(),
				3072,
				Tiers.IRON.getSpeed() * 1.1f,
				3.5f,
				16,
				BlockTags.NEEDS_IRON_TOOL,
				() -> component
			);
		}

		public static final SwordItem TRITANIUM_SWORD = new SwordItem(TRITANIUM_COMPONENT, 4, -2.7f, new Item.Properties().tab(OverdriveThatMatters.CREATIVE_TAB));
		public static final ShovelItem TRITANIUM_SHOVEL = new ShovelItem(TRITANIUM_COMPONENT, 1.5f, -2.4f, new Item.Properties().tab(OverdriveThatMatters.CREATIVE_TAB));
		public static final AxeItem TRITANIUM_AXE = new AxeItem(TRITANIUM_COMPONENT, 8.5f, -3.4f, new Item.Properties().tab(OverdriveThatMatters.CREATIVE_TAB));
		public static final PickaxeItem TRITANIUM_PICKAXE = new PickaxeItem(TRITANIUM_COMPONENT, 2, -2.8f, new Item.Properties().tab(OverdriveThatMatters.CREATIVE_TAB));
		public static final HoeItem TRITANIUM_HOE = new HoeItem(TRITANIUM_COMPONENT, 0, -3.4f, new Item.Properties().tab(OverdriveThatMatters.CREATIVE_TAB));

		public static final Item[] TRITANIUM_TOOLS = new Item[] {
			TRITANIUM_SWORD,
			TRITANIUM_SHOVEL,
			TRITANIUM_AXE,
			TRITANIUM_PICKAXE,
			TRITANIUM_HOE
		};

		public static final ItemTritaniumArmor TRITANIUM_HELMET = new ItemTritaniumArmor(EquipmentSlot.HEAD);
		public static final ItemTritaniumArmor TRITANIUM_CHESTPLATE = new ItemTritaniumArmor(EquipmentSlot.CHEST);
		public static final ItemTritaniumArmor TRITANIUM_PANTS = new ItemTritaniumArmor(EquipmentSlot.LEGS);
		public static final ItemTritaniumArmor TRITANIUM_BOOTS = new ItemTritaniumArmor(EquipmentSlot.FEET);

		public static final Item[] TRITANIUM_ARMOR = new Item[] {
				TRITANIUM_HELMET,
				TRITANIUM_CHESTPLATE,
				TRITANIUM_PANTS,
				TRITANIUM_BOOTS
		};

		public static final Item MATTER_IO_PORT = new Item(new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));
		public static final Item MATTER_TRANSFORM_MATRIX = new Item(new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));
		public static final Item ENERGY_BUS = new Item(new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));
		public static final Item ELECTRIC_PARTS = new Item(new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));
		public static final Item MACHINE_FRAME = new Item(new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));
		public static final Item TRITANIUM_PLATE = new Item(new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));
		public static final Item IRON_PLATE = new Item(new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));
		public static final Item COPPER_WIRING = new Item(new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));
		public static final Item GOLD_WIRING = new Item(new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));
		public static final Item PORTABLE_CONDENSATION_DRIVE_CASING = new Item(new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));
		public static final Item PORTABLE_DENSE_CONDENSATION_DRIVE_CASING = new Item(new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));
		public static final Item CIRCUIT_PLATING = new Item(new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));
		public static final Item BASIC_CONTROL_CIRCUIT = new Item(new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));
		public static final Item ADVANCED_CONTROL_CIRCUIT = new Item(new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));

		public static final Item BLACK_HOLE_SCANNER = new Item(new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB)) {
			@Override
			public void appendHoverText(@Nonnull ItemStack p_40572_, @Nullable Level p_40573_, @Nonnull List<Component> p_40574_, @Nonnull TooltipFlag p_40575_) {
				super.appendHoverText(p_40572_, p_40573_, p_40574_, p_40575_);

				p_40574_.add(new TranslatableComponent("item.overdrive_that_matters.black_hole_scanner.desc").withStyle(ChatFormatting.GRAY));
				p_40574_.add(new TranslatableComponent("item.overdrive_that_matters.black_hole_scanner.desc2").withStyle(ChatFormatting.DARK_GRAY));
			}
		};
		public static final Item GRAVITATION_FIELD_LIMITER = new Item(new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));
		public static final Item GRAVITATION_FIELD_SENSOR = new Item(new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));
		public static final ItemPortableGravitationStabilizer PORTABLE_GRAVITATION_STABILIZER = new ItemPortableGravitationStabilizer();

		public static final BlockItem BLACK_HOLE = new BlockItem(Blocks.BLACK_HOLE, new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));

		public static final ItemGravitationalDisruptor GRAVITATIONAL_DISRUPTOR = new ItemGravitationalDisruptor();

		public static final ItemPill PILL_ANDROID = new ItemPill(PillType.BECOME_ANDROID);
		public static final ItemPill PILL_HUMANE = new ItemPill(PillType.BECOME_HUMANE);
		public static final ItemPill PILL_OBLIVION = new ItemPill(PillType.OBLIVION);
		public static final ItemPillHeal PILL_HEAL = new ItemPillHeal();

		public static final ItemBattery BATTERY_CRUDE = new ItemBattery(new Fraction(30_000), new Fraction(150), new Fraction(150));
		public static final ItemBattery BATTERY_BASIC = new ItemBattery(new Fraction(60_000), new Fraction(300), new Fraction(300));
		public static final ItemBattery BATTERY_NORMAL = new ItemBattery(new Fraction(250_000), new Fraction(1000), new Fraction(1000));
		public static final ItemBattery BATTERY_DENSE = new ItemBattery(new Fraction(1_000_000), new Fraction(2000), new Fraction(2000));
		public static final ItemBattery BATTERY_CAPACITOR = new ItemBattery(new Fraction(150_000), new Fraction(15000), new Fraction(15000));
		public static final ItemBattery BATTERY_CREATIVE = new ItemBattery();

		public static final ItemBattery[] BATTERIES = {
			BATTERY_CRUDE,
			BATTERY_BASIC,
			BATTERY_NORMAL,
			BATTERY_DENSE,
			BATTERY_CAPACITOR,
		};

		public static final Item MATTER_CAPACITOR_PARTS = new Item(new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));

		public static final Item[] DATAGEN_COMPONENTS = {
			ENERGY_BUS,
			ELECTRIC_PARTS,
			TRITANIUM_PLATE,
			IRON_PLATE,
			COPPER_WIRING,
			GOLD_WIRING,
			CIRCUIT_PLATING,
			BASIC_CONTROL_CIRCUIT,
			ADVANCED_CONTROL_CIRCUIT,
			MATTER_CAPACITOR_PARTS,
			MATTER_IO_PORT,
			MATTER_TRANSFORM_MATRIX,
		};

		public static final ItemMatterCapacitor MATTER_CAPACITOR_BASIC = new ItemMatterCapacitor(new Fraction("4"));
		public static final ItemMatterCapacitor MATTER_CAPACITOR_NORMAL = new ItemMatterCapacitor(new Fraction("10"));
		public static final ItemMatterCapacitor MATTER_CAPACITOR_DENSE = new ItemMatterCapacitor(new Fraction("40"));
		public static final ItemMatterCapacitor MATTER_CAPACITOR_CREATIVE = new ItemMatterCapacitor();

		public static final ItemPatternStorage PATTERN_DRIVE_NORMAL = new ItemPatternStorage(4);
		public static final ItemPatternStorage PATTERN_DRIVE_CREATIVE = new ItemPatternStorage();

		public static final ItemPortableCondensationDrive PORTABLE_CONDENSATION_DRIVE = new ItemPortableCondensationDrive(4000);
		public static final ItemPortableCondensationDrive PORTABLE_DENSE_CONDENSATION_DRIVE = new ItemPortableCondensationDrive(25000);

		public static final Item NUTRIENT_PASTE = new Item(new Item.Properties().stacksTo(64).food(new FoodProperties.Builder().meat().nutrition(8).saturationMod(0.9F).build()).tab(OverdriveThatMatters.CREATIVE_TAB));

		public static final Item[] CRATES = new Item[Registry.CRATES.length];

		public static final Item TRITANIUM_BLOCK = new BlockItem(Blocks.TRITANIUM_BLOCK, new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));
		public static final Item TRITANIUM_STRIPED_BLOCK = new BlockItem(Blocks.TRITANIUM_STRIPED_BLOCK, new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));
		public static final Item CARBON_FIBRE_BLOCK = new BlockItem(Blocks.CARBON_FIBRE_BLOCK, new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));

		static {
			for (int i = 0; i < Registry.CRATES.length; i++) {
				CRATES[i] = new BlockItem(Blocks.CRATES[i], new Item.Properties().stacksTo(64).tab(OverdriveThatMatters.CREATIVE_TAB));
				CRATES[i].setRegistryName(Registry.CRATES[i].name);
			}

			ANDROID_STATION.setRegistryName(Names.ANDROID_STATION);
			BATTERY_BANK.setRegistryName(Names.BATTERY_BANK);
			MATTER_DECOMPOSER.setRegistryName(Names.MATTER_DECOMPOSER);
			MATTER_CAPACITOR_BANK.setRegistryName(Names.MATTER_CAPACITOR_BANK);
			MATTER_CABLE.setRegistryName(Names.MATTER_CABLE);
			PATTERN_STORAGE.setRegistryName(Names.PATTERN_STORAGE);
			MATTER_SCANNER.setRegistryName(Names.MATTER_SCANNER);
			MATTER_PANEL.setRegistryName(Names.MATTER_PANEL);
			MATTER_REPLICATOR.setRegistryName(Names.MATTER_REPLICATOR);
			MATTER_BOTTLER.setRegistryName(Names.MATTER_BOTTLER);
			DRIVE_VIEWER.setRegistryName(Names.DRIVE_VIEWER);
			BLACK_HOLE.setRegistryName(Names.BLACK_HOLE);
			CARGO_CRATE.setRegistryName(Names.CARGO_CRATE);
			DRIVE_RACK.setRegistryName(Names.DRIVE_RACK);
			ITEM_MONITOR.setRegistryName(Names.ITEM_MONITOR);
			ENERGY_COUNTER.setRegistryName(Names.ENERGY_COUNTER);
			CHEMICAL_GENERATOR.setRegistryName(Names.CHEMICAL_GENERATOR);
			PLATE_PRESS.setRegistryName(Names.PLATE_PRESS);
			GRAVITATION_STABILIZER.setRegistryName(Names.GRAVITATION_STABILIZER);

			DEBUG_EXPLOSION_SMALL.setRegistryName(Names.DEBUG_EXPLOSION_SMALL);
			DEBUG_SPHERE_POINTS.setRegistryName(Names.DEBUG_SPHERE_POINTS);

			TRITANIUM_ORE.setRegistryName(Names.TRITANIUM_ORE);
			DEEPSLATE_TRITANIUM_ORE.setRegistryName(Names.DEEPSLATE_TRITANIUM_ORE);
			TRITANIUM_RAW_BLOCK.setRegistryName(Names.TRITANIUM_RAW_BLOCK);
			TRITANIUM_ORE_CLUMP.setRegistryName(Names.TRITANIUM_ORE_CLUMP);
			TRITANIUM_INGOT.setRegistryName(Names.TRITANIUM_INGOT);
			MATTER_IO_PORT.setRegistryName(Names.MATTER_IO_PORT);
			MATTER_TRANSFORM_MATRIX.setRegistryName(Names.MATTER_TRANSFORM_MATRIX);
			ENERGY_BUS.setRegistryName(Names.ENERGY_BUS);
			ELECTRIC_PARTS.setRegistryName(Names.ELECTRIC_PARTS);
			MACHINE_FRAME.setRegistryName(Names.MACHINE_FRAME);
			TRITANIUM_PLATE.setRegistryName(Names.TRITANIUM_PLATE);
			IRON_PLATE.setRegistryName(Names.IRON_PLATE);
			COPPER_WIRING.setRegistryName(Names.COPPER_WIRING);
			GOLD_WIRING.setRegistryName(Names.GOLD_WIRING);
			PORTABLE_CONDENSATION_DRIVE_CASING.setRegistryName(Names.PORTABLE_CONDENSATION_DRIVE_CASING);
			PORTABLE_DENSE_CONDENSATION_DRIVE_CASING.setRegistryName(Names.PORTABLE_DENSE_CONDENSATION_DRIVE_CASING);
			CIRCUIT_PLATING.setRegistryName(Names.CIRCUIT_PLATING);
			BASIC_CONTROL_CIRCUIT.setRegistryName(Names.BASIC_CONTROL_CIRCUIT);
			ADVANCED_CONTROL_CIRCUIT.setRegistryName(Names.ADVANCED_CONTROL_CIRCUIT);

			BLACK_HOLE_SCANNER.setRegistryName(Names.BLACK_HOLE_SCANNER);
			GRAVITATION_FIELD_LIMITER.setRegistryName(Names.GRAVITATION_FIELD_LIMITER);
			GRAVITATION_FIELD_SENSOR.setRegistryName(Names.GRAVITATION_FIELD_SENSOR);
			PORTABLE_GRAVITATION_STABILIZER.setRegistryName(Names.PORTABLE_GRAVITATION_STABILIZER);

			GRAVITATIONAL_DISRUPTOR.setRegistryName(Names.GRAVITATIONAL_DISRUPTOR);
			PILL_ANDROID.setRegistryName(Names.PILL_ANDROID);
			PILL_HUMANE.setRegistryName(Names.PILL_HUMANE);
			PILL_OBLIVION.setRegistryName(Names.PILL_OBLIVION);
			PILL_HEAL.setRegistryName(Names.PILL_HEAL);
			BATTERY_CRUDE.setRegistryName(Names.BATTERY_CRUDE);
			BATTERY_BASIC.setRegistryName(Names.BATTERY_BASIC);
			BATTERY_NORMAL.setRegistryName(Names.BATTERY_NORMAL);
			BATTERY_DENSE.setRegistryName(Names.BATTERY_DENSE);
			BATTERY_CAPACITOR.setRegistryName(Names.BATTERY_CAPACITOR);
			BATTERY_CREATIVE.setRegistryName(Names.BATTERY_CREATIVE);

			MATTER_CAPACITOR_PARTS.setRegistryName(Names.MATTER_CAPACITOR_PARTS);
			MATTER_CAPACITOR_BASIC.setRegistryName(Names.MATTER_CAPACITOR_BASIC);
			MATTER_CAPACITOR_NORMAL.setRegistryName(Names.MATTER_CAPACITOR_NORMAL);
			MATTER_CAPACITOR_DENSE.setRegistryName(Names.MATTER_CAPACITOR_DENSE);
			MATTER_CAPACITOR_CREATIVE.setRegistryName(Names.MATTER_CAPACITOR_CREATIVE);

			PATTERN_DRIVE_NORMAL.setRegistryName(Names.PATTERN_DRIVE_NORMAL);
			PATTERN_DRIVE_CREATIVE.setRegistryName(Names.PATTERN_DRIVE_CREATIVE);
			PORTABLE_CONDENSATION_DRIVE.setRegistryName(Names.PORTABLE_CONDENSATION_DRIVE);
			PORTABLE_DENSE_CONDENSATION_DRIVE.setRegistryName(Names.PORTABLE_DENSE_CONDENSATION_DRIVE);

			NUTRIENT_PASTE.setRegistryName(Names.NUTRIENT_PASTE);

			TRITANIUM_BLOCK.setRegistryName(Names.TRITANIUM_BLOCK);
			TRITANIUM_STRIPED_BLOCK.setRegistryName(Names.TRITANIUM_STRIPED_BLOCK);
			CARBON_FIBRE_BLOCK.setRegistryName(Names.CARBON_FIBRE_BLOCK);

			TRITANIUM_SWORD.setRegistryName(Names.TRITANIUM_SWORD);
			TRITANIUM_PICKAXE.setRegistryName(Names.TRITANIUM_PICKAXE);
			TRITANIUM_SHOVEL.setRegistryName(Names.TRITANIUM_SHOVEL);
			TRITANIUM_AXE.setRegistryName(Names.TRITANIUM_AXE);
			TRITANIUM_HOE.setRegistryName(Names.TRITANIUM_HOE);

			TRITANIUM_HELMET.setRegistryName(Names.TRITANIUM_HELMET);
			TRITANIUM_CHESTPLATE.setRegistryName(Names.TRITANIUM_CHESTPLATE);
			TRITANIUM_PANTS.setRegistryName(Names.TRITANIUM_PANTS);
			TRITANIUM_BOOTS.setRegistryName(Names.TRITANIUM_BOOTS);
		}

		@SubscribeEvent
		@SuppressWarnings("unused")
		public static void register(final RegistryEvent.Register<Item> event) {
			event.getRegistry().register(ANDROID_STATION);
			event.getRegistry().register(BATTERY_BANK);
			event.getRegistry().register(MATTER_DECOMPOSER);
			event.getRegistry().register(MATTER_CAPACITOR_BANK);
			event.getRegistry().register(MATTER_CABLE);
			event.getRegistry().register(PATTERN_STORAGE);
			event.getRegistry().register(MATTER_SCANNER);
			event.getRegistry().register(MATTER_PANEL);
			event.getRegistry().register(MATTER_REPLICATOR);
			event.getRegistry().register(MATTER_BOTTLER);
			event.getRegistry().register(DRIVE_VIEWER);
			event.getRegistry().register(BLACK_HOLE);
			event.getRegistry().register(CARGO_CRATE);
			event.getRegistry().register(DRIVE_RACK);
			event.getRegistry().register(ITEM_MONITOR);
			event.getRegistry().register(ENERGY_COUNTER);
			event.getRegistry().register(CHEMICAL_GENERATOR);
			event.getRegistry().register(PLATE_PRESS);
			event.getRegistry().register(GRAVITATION_STABILIZER);

			event.getRegistry().register(DEBUG_EXPLOSION_SMALL);
			event.getRegistry().register(DEBUG_SPHERE_POINTS);

			event.getRegistry().register(TRITANIUM_ORE);
			event.getRegistry().register(DEEPSLATE_TRITANIUM_ORE);
			event.getRegistry().register(TRITANIUM_RAW_BLOCK);

			event.getRegistry().register(TRITANIUM_ORE_CLUMP);
			event.getRegistry().register(TRITANIUM_INGOT);
			event.getRegistry().register(MATTER_IO_PORT);
			event.getRegistry().register(MATTER_TRANSFORM_MATRIX);
			event.getRegistry().register(ENERGY_BUS);
			event.getRegistry().register(ELECTRIC_PARTS);
			event.getRegistry().register(MACHINE_FRAME);
			event.getRegistry().register(TRITANIUM_PLATE);
			event.getRegistry().register(IRON_PLATE);
			event.getRegistry().register(COPPER_WIRING);
			event.getRegistry().register(GOLD_WIRING);
			event.getRegistry().register(PORTABLE_CONDENSATION_DRIVE_CASING);
			event.getRegistry().register(PORTABLE_DENSE_CONDENSATION_DRIVE_CASING);
			event.getRegistry().register(CIRCUIT_PLATING);
			event.getRegistry().register(BASIC_CONTROL_CIRCUIT);
			event.getRegistry().register(ADVANCED_CONTROL_CIRCUIT);

			event.getRegistry().register(BLACK_HOLE_SCANNER);
			event.getRegistry().register(GRAVITATION_FIELD_LIMITER);
			event.getRegistry().register(GRAVITATION_FIELD_SENSOR);
			event.getRegistry().register(PORTABLE_GRAVITATION_STABILIZER);

			event.getRegistry().register(GRAVITATIONAL_DISRUPTOR);
			event.getRegistry().register(PILL_ANDROID);
			event.getRegistry().register(PILL_HUMANE);
			event.getRegistry().register(PILL_OBLIVION);
			event.getRegistry().register(PILL_HEAL);
			event.getRegistry().register(BATTERY_CRUDE);
			event.getRegistry().register(BATTERY_BASIC);
			event.getRegistry().register(BATTERY_NORMAL);
			event.getRegistry().register(BATTERY_DENSE);
			event.getRegistry().register(BATTERY_CAPACITOR);
			event.getRegistry().register(BATTERY_CREATIVE);

			event.getRegistry().register(MATTER_CAPACITOR_PARTS);
			event.getRegistry().register(MATTER_CAPACITOR_BASIC);
			event.getRegistry().register(MATTER_CAPACITOR_NORMAL);
			event.getRegistry().register(MATTER_CAPACITOR_DENSE);
			event.getRegistry().register(MATTER_CAPACITOR_CREATIVE);

			event.getRegistry().register(PATTERN_DRIVE_NORMAL);
			event.getRegistry().register(PATTERN_DRIVE_CREATIVE);
			event.getRegistry().register(PORTABLE_CONDENSATION_DRIVE);
			event.getRegistry().register(PORTABLE_DENSE_CONDENSATION_DRIVE);

			event.getRegistry().register(NUTRIENT_PASTE);

			event.getRegistry().register(TRITANIUM_BLOCK);
			event.getRegistry().register(TRITANIUM_STRIPED_BLOCK);
			event.getRegistry().register(CARBON_FIBRE_BLOCK);

			for (var crate : CRATES) {
				event.getRegistry().register(crate);
			}

			event.getRegistry().register(TRITANIUM_SWORD);
			event.getRegistry().register(TRITANIUM_SHOVEL);
			event.getRegistry().register(TRITANIUM_AXE);
			event.getRegistry().register(TRITANIUM_PICKAXE);
			event.getRegistry().register(TRITANIUM_HOE);

			event.getRegistry().register(TRITANIUM_HELMET);
			event.getRegistry().register(TRITANIUM_CHESTPLATE);
			event.getRegistry().register(TRITANIUM_PANTS);
			event.getRegistry().register(TRITANIUM_BOOTS);

			// OverdriveThatMatters.LOGGER.info("Registered items");
		}
	}

	public static class BlockEntities {
		public static final BlockEntityType<BlockEntityAndroidStation> ANDROID_STATION = BlockEntityType.Builder.of(BlockEntityAndroidStation::new, Blocks.ANDROID_STATION).build(null);
		public static final BlockEntityType<BlockEntityBatteryBank> BATTERY_BANK = BlockEntityType.Builder.of(BlockEntityBatteryBank::new, Blocks.BATTERY_BANK).build(null);
		public static final BlockEntityType<BlockEntityMatterDecomposer> MATTER_DECOMPOSER = BlockEntityType.Builder.of(BlockEntityMatterDecomposer::new, Blocks.MATTER_DECOMPOSER).build(null);
		public static final BlockEntityType<BlockEntityMatterCapacitorBank> MATTER_CAPACITOR_BANK = BlockEntityType.Builder.of(BlockEntityMatterCapacitorBank::new, Blocks.MATTER_CAPACITOR_BANK).build(null);
		public static final BlockEntityType<BlockEntityMatterCable> MATTER_CABLE = BlockEntityType.Builder.of(BlockEntityMatterCable::new, Blocks.MATTER_CABLE).build(null);
		public static final BlockEntityType<BlockEntityPatternStorage> PATTERN_STORAGE = BlockEntityType.Builder.of(BlockEntityPatternStorage::new, Blocks.PATTERN_STORAGE).build(null);
		public static final BlockEntityType<BlockEntityMatterScanner> MATTER_SCANNER = BlockEntityType.Builder.of(BlockEntityMatterScanner::new, Blocks.MATTER_SCANNER).build(null);
		public static final BlockEntityType<BlockEntityMatterPanel> MATTER_PANEL = BlockEntityType.Builder.of(BlockEntityMatterPanel::new, Blocks.MATTER_PANEL).build(null);
		public static final BlockEntityType<BlockEntityMatterReplicator> MATTER_REPLICATOR = BlockEntityType.Builder.of(BlockEntityMatterReplicator::new, Blocks.MATTER_REPLICATOR).build(null);
		public static final BlockEntityType<BlockEntityMatterBottler> MATTER_BOTTLER = BlockEntityType.Builder.of(BlockEntityMatterBottler::new, Blocks.MATTER_BOTTLER).build(null);
		public static final BlockEntityType<BlockEntityDriveViewer> DRIVE_VIEWER = BlockEntityType.Builder.of(BlockEntityDriveViewer::new, Blocks.DRIVE_VIEWER).build(null);
		public static final BlockEntityType<BlockEntityBlackHole> BLACK_HOLE = BlockEntityType.Builder.of(BlockEntityBlackHole::new, Blocks.BLACK_HOLE).build(null);
		public static final BlockEntityType<BlockEntityCargoCrate> CARGO_CRATE = BlockEntityType.Builder.of(BlockEntityCargoCrate::new, Blocks.CARGO_CRATE).build(null);
		public static final BlockEntityType<BlockEntityDriveRack> DRIVE_RACK = BlockEntityType.Builder.of(BlockEntityDriveRack::new, Blocks.DRIVE_RACK).build(null);
		public static final BlockEntityType<BlockEntityItemMonitor> ITEM_MONITOR = BlockEntityType.Builder.of(BlockEntityItemMonitor::new, Blocks.ITEM_MONITOR).build(null);
		public static final BlockEntityType<BlockEntityEnergyCounter> ENERGY_COUNTER = BlockEntityType.Builder.of(BlockEntityEnergyCounter::new, Blocks.ENERGY_COUNTER).build(null);
		public static final BlockEntityType<BlockEntityChemicalGenerator> CHEMICAL_GENERATOR = BlockEntityType.Builder.of(BlockEntityChemicalGenerator::new, Blocks.CHEMICAL_GENERATOR).build(null);
		public static final BlockEntityType<BlockEntityPlatePress> PLATE_PRESS = BlockEntityType.Builder.of(BlockEntityPlatePress::new, Blocks.PLATE_PRESS).build(null);
		public static final BlockEntityType<BlockEntityGravitationStabilizer> GRAVITATION_STABILIZER = BlockEntityType.Builder.of(BlockEntityGravitationStabilizer::new, Blocks.GRAVITATION_STABILIZER).build(null);

		public static final BlockEntityType<BlockEntityExplosionDebugger> DEBUG_EXPLOSION_SMALL = BlockEntityType.Builder.of(BlockEntityExplosionDebugger::new, Blocks.DEBUG_EXPLOSION_SMALL).build(null);
		public static final BlockEntityType<BlockEntitySphereDebugger> DEBUG_SPHERE_POINTS = BlockEntityType.Builder.of(BlockEntitySphereDebugger::new, Blocks.DEBUG_SPHERE_POINTS).build(null);

		static {
			ANDROID_STATION.setRegistryName(Names.ANDROID_STATION);
			BATTERY_BANK.setRegistryName(Names.BATTERY_BANK);
			MATTER_DECOMPOSER.setRegistryName(Names.MATTER_DECOMPOSER);
			MATTER_CAPACITOR_BANK.setRegistryName(Names.MATTER_CAPACITOR_BANK);
			MATTER_CABLE.setRegistryName(Names.MATTER_CABLE);
			PATTERN_STORAGE.setRegistryName(Names.PATTERN_STORAGE);
			MATTER_SCANNER.setRegistryName(Names.MATTER_SCANNER);
			MATTER_PANEL.setRegistryName(Names.MATTER_PANEL);
			MATTER_REPLICATOR.setRegistryName(Names.MATTER_REPLICATOR);
			MATTER_BOTTLER.setRegistryName(Names.MATTER_BOTTLER);
			DRIVE_VIEWER.setRegistryName(Names.DRIVE_VIEWER);
			BLACK_HOLE.setRegistryName(Names.BLACK_HOLE);
			CARGO_CRATE.setRegistryName(Names.CARGO_CRATE);
			DRIVE_RACK.setRegistryName(Names.DRIVE_RACK);
			ITEM_MONITOR.setRegistryName(Names.ITEM_MONITOR);
			ENERGY_COUNTER.setRegistryName(Names.ENERGY_COUNTER);
			CHEMICAL_GENERATOR.setRegistryName(Names.CHEMICAL_GENERATOR);
			PLATE_PRESS.setRegistryName(Names.PLATE_PRESS);
			GRAVITATION_STABILIZER.setRegistryName(Names.GRAVITATION_STABILIZER);

			DEBUG_EXPLOSION_SMALL.setRegistryName(Names.DEBUG_EXPLOSION_SMALL);
			DEBUG_SPHERE_POINTS.setRegistryName(Names.DEBUG_SPHERE_POINTS);
		}

		@SubscribeEvent
		@SuppressWarnings("unused")
		public static void register(final RegistryEvent.Register<BlockEntityType<?>> event) {
			event.getRegistry().register(ANDROID_STATION);
			event.getRegistry().register(BATTERY_BANK);
			event.getRegistry().register(MATTER_DECOMPOSER);
			event.getRegistry().register(MATTER_CAPACITOR_BANK);
			event.getRegistry().register(MATTER_CABLE);
			event.getRegistry().register(PATTERN_STORAGE);
			event.getRegistry().register(MATTER_SCANNER);
			event.getRegistry().register(MATTER_PANEL);
			event.getRegistry().register(MATTER_REPLICATOR);
			event.getRegistry().register(MATTER_BOTTLER);
			event.getRegistry().register(DRIVE_VIEWER);
			event.getRegistry().register(BLACK_HOLE);
			event.getRegistry().register(CARGO_CRATE);
			event.getRegistry().register(DRIVE_RACK);
			event.getRegistry().register(ITEM_MONITOR);
			event.getRegistry().register(ENERGY_COUNTER);
			event.getRegistry().register(CHEMICAL_GENERATOR);
			event.getRegistry().register(PLATE_PRESS);
			event.getRegistry().register(GRAVITATION_STABILIZER);

			event.getRegistry().register(DEBUG_EXPLOSION_SMALL);
			event.getRegistry().register(DEBUG_SPHERE_POINTS);

			// OverdriveThatMatters.LOGGER.info("Registered block entities");
		}


		@SubscribeEvent
		@SuppressWarnings("unused")
		public static void registerRenderers(final FMLClientSetupEvent event) {
			BlockEntityRenderers.register(BLACK_HOLE, BlackHoleRenderer::new);
			BlockEntityRenderers.register(GRAVITATION_STABILIZER, GravitationStabilizerRenderer::new);
		}
	}

	public static class AndroidFeatures {
		public static final AndroidFeatureType<AndroidFeature> AIR_BAGS = new AndroidFeatureType<>(AndroidFeature::new);
		public static final AndroidFeatureType<AndroidLimbOverclocking> LIMB_OVERCLOCKING = new AndroidFeatureType<>(AndroidLimbOverclocking::new);
		public static final AndroidFeatureType<AndroidNanobotsRegeneration> NANOBOTS_REGENERATION = new AndroidFeatureType<>(AndroidNanobotsRegeneration::new);
		public static final AndroidFeatureType<AndroidNanobotsArmor> NANOBOTS_ARMOR = new AndroidFeatureType<>(AndroidNanobotsArmor::new);
		public static final AndroidFeatureType<AndroidExtendedReach> EXTENDED_REACH = new AndroidFeatureType<>(AndroidExtendedReach::new);

		static {
			AIR_BAGS.setRegistryName(Names.AIR_BAGS);
			LIMB_OVERCLOCKING.setRegistryName(Names.LIMB_OVERCLOCKING);
			NANOBOTS_REGENERATION.setRegistryName(Names.NANOBOTS_REGENERATION);
			NANOBOTS_ARMOR.setRegistryName(Names.NANOBOTS_ARMOR);
			EXTENDED_REACH.setRegistryName(Names.EXTENDED_REACH);
		}

		@SubscribeEvent
		@SuppressWarnings("unused")
		public static void register(final RegistryEvent.Register<AndroidFeatureType<?>> event) {
			event.getRegistry().register(AIR_BAGS);
			event.getRegistry().register(LIMB_OVERCLOCKING);
			event.getRegistry().register(NANOBOTS_REGENERATION);
			event.getRegistry().register(NANOBOTS_ARMOR);
			event.getRegistry().register(EXTENDED_REACH);
		}
	}

	public static class AndroidResearch {
		public static final ResourceLocation ICONS = new ResourceLocation(OverdriveThatMatters.MOD_ID, "textures/gui/android_upgrades.png");
		public static final SkinElement ICON_TRANSFER;
		public static final SkinElement ICON_ATTACK_BOOST;
		public static final SkinElement ICON_PLASMA_SHIELD_BOOST;
		public static final SkinElement ICON_CLOAK;
		public static final SkinElement ICON_GRAVITATIONAL_STABILIZER;
		public static final SkinElement ICON_AIR_BAGS;
		public static final SkinElement ICON_JUMP_BOOST;

		public static final SkinElement ICON_FEATHER_FALLING;
		public static final SkinElement ICON_ARC;
		public static final SkinElement ICON_ARROW;
		public static final SkinElement ICON_ARMOR;
		public static final SkinElement ICON_NANOBOTS;
		public static final SkinElement ICON_NIGHT_VISION;
		public static final SkinElement ICON_OXYGEN_SUPPLY;

		public static final SkinElement ICON_PLASMA_SHIELD;
		public static final SkinElement ICON_SHOCKWAVE;
		public static final SkinElement ICON_LIMB_OVERCLOCKING;
		public static final SkinElement ICON_STEP_ASSIST;
		public static final SkinElement ICON_ENDER_TELEPORT;
		public static final SkinElement ICON_WIRELESS_CHARGING;
		public static final SkinElement ICON_UNKNOWN;

		public static final SkinElement ICON_EXTENDED_REACH;

		static {
			int x = 0;
			int y = 0;

			ICON_TRANSFER = new SkinElement(ICONS, x, y, 18, 18, 126, 126); x += 18;
			ICON_ATTACK_BOOST = new SkinElement(ICONS, x, y, 18, 18, 126, 126); x += 18;
			ICON_PLASMA_SHIELD_BOOST = new SkinElement(ICONS, x, y, 18, 18, 126, 126); x += 18;
			ICON_CLOAK = new SkinElement(ICONS, x, y, 18, 18, 126, 126); x += 18;
			ICON_GRAVITATIONAL_STABILIZER = new SkinElement(ICONS, x, y, 18, 18, 126, 126); x += 18;
			ICON_AIR_BAGS = new SkinElement(ICONS, x, y, 18, 18, 126, 126); x += 18;
			ICON_JUMP_BOOST = new SkinElement(ICONS, x, y, 18, 18, 126, 126);

			y += 18;
			x = 0;

			ICON_FEATHER_FALLING = new SkinElement(ICONS, x, y, 18, 18, 126, 126); x += 18;
			ICON_ARC = new SkinElement(ICONS, x, y, 18, 18, 126, 126); x += 18;
			ICON_ARROW = new SkinElement(ICONS, x, y, 18, 18, 126, 126); x += 18;
			ICON_ARMOR = new SkinElement(ICONS, x, y, 18, 18, 126, 126); x += 18;
			ICON_NANOBOTS = new SkinElement(ICONS, x, y, 18, 18, 126, 126); x += 18;
			ICON_NIGHT_VISION = new SkinElement(ICONS, x, y, 18, 18, 126, 126); x += 18;
			ICON_OXYGEN_SUPPLY = new SkinElement(ICONS, x, y, 18, 18, 126, 126);

			y += 18;
			x = 0;

			ICON_PLASMA_SHIELD = new SkinElement(ICONS, x, y, 18, 18, 126, 126); x += 18;
			ICON_SHOCKWAVE = new SkinElement(ICONS, x, y, 18, 18, 126, 126); x += 18;
			ICON_LIMB_OVERCLOCKING = new SkinElement(ICONS, x, y, 18, 18, 126, 126); x += 18;
			ICON_STEP_ASSIST = new SkinElement(ICONS, x, y, 18, 18, 126, 126); x += 18;
			ICON_ENDER_TELEPORT = new SkinElement(ICONS, x, y, 18, 18, 126, 126); x += 18;
			ICON_WIRELESS_CHARGING = new SkinElement(ICONS, x, y, 18, 18, 126, 126); x += 18;
			ICON_UNKNOWN = new SkinElement(ICONS, x, y, 18, 18, 126, 126);

			y += 18;
			x = 0;

			ICON_EXTENDED_REACH = new SkinElement(ICONS, x, y, 18, 18, 126, 126); x += 18;
		}

		public static final AndroidResearchType<ru.dbotthepony.mc.otm.android.AndroidResearch> AIR_BAGS =
			new AndroidResearchBuilder()
				.setExperienceCost(18)
				.addFeatureResult(AndroidFeatures.AIR_BAGS)
				.withDescription()
				.withIcon(ICON_AIR_BAGS)
				.build();

		public static final AndroidResearchType<ru.dbotthepony.mc.otm.android.AndroidResearch> EXTENDED_REACH =
			new AndroidResearchBuilder()
				.setExperienceCost(40)
				.addFeatureResult(AndroidFeatures.EXTENDED_REACH)
				.withDescription()
				.withIcon(ICON_EXTENDED_REACH)
				.build();

		public static final AndroidResearchType<ru.dbotthepony.mc.otm.android.AndroidResearch>[] LIMB_OVERCLOCKING = new AndroidResearchType[4];
		public static final AndroidResearchType<ru.dbotthepony.mc.otm.android.AndroidResearch>[] NANOBOTS_REGENERATION = new AndroidResearchType[4];
		public static final AndroidResearchType<ru.dbotthepony.mc.otm.android.AndroidResearch>[] NANOBOTS_ARMOR_STRENGTH = new AndroidResearchType[3];
		public static final AndroidResearchType<ru.dbotthepony.mc.otm.android.AndroidResearch>[] NANOBOTS_ARMOR_SPEED = new AndroidResearchType[3];

		public static final AndroidResearchType<ru.dbotthepony.mc.otm.android.AndroidResearch> NANOBOTS =
			new AndroidResearchBuilder()
				.setExperienceCost(15)
				.withDescription()
				.withIcon(ICON_NANOBOTS)
				.build();

		public static final AndroidResearchType<ru.dbotthepony.mc.otm.android.AndroidResearch> NANOBOTS_ARMOR =
			new AndroidResearchBuilder()
				.setExperienceCost(25)
				.withDescription()
				.addPrerequisite(Names.NANOBOTS)
				.addFeatureResult(Names.NANOBOTS_ARMOR)
				.withIcon(ICON_ARMOR)
				.build();

		static {
			AIR_BAGS.setRegistryName(Names.AIR_BAGS);
			NANOBOTS.setRegistryName(Names.NANOBOTS);
			NANOBOTS_ARMOR.setRegistryName(Names.NANOBOTS_ARMOR);
			EXTENDED_REACH.setRegistryName(Names.EXTENDED_REACH);

			for (int i = 0; i < 4; i++) {
				var limbs = new AndroidResearchBuilder()
					.setExperienceCost(18 + i * 8)
					.withIconText(new TextComponent(String.valueOf(i + 1)))
					.withIcon(ICON_LIMB_OVERCLOCKING)
					.withName(new TranslatableComponent("android_research.overdrive_that_matters.limb_overclocking", i + 1))
					.withDescription(new TranslatableComponent("android_research.overdrive_that_matters.limb_overclocking.description", (i + 1) * 8, (i + 1) * 6))
					.addFeatureResult(Names.LIMB_OVERCLOCKING, i);

				var regeneration = new AndroidResearchBuilder()
					.setExperienceCost(20 + i * 6)
					.withIconText(new TextComponent(String.valueOf(i + 1)))
					.withIcon(ICON_NANOBOTS)
					.withName(new TranslatableComponent("android_research.overdrive_that_matters.nanobots_regeneration", i + 1))
					.withDescription(
						i > 0 ? new TranslatableComponent("android_research.overdrive_that_matters.nanobots_regeneration.description_improve") : new TranslatableComponent("android_research.overdrive_that_matters.nanobots_regeneration.description"))
					.addFeatureResult(Names.NANOBOTS_REGENERATION, i);

				if (i > 0) {
					limbs.addPrerequisite(Names.LIMB_OVERCLOCKING_LIST[i - 1]);
					regeneration.addPrerequisite(Names.NANOBOTS_REGENERATION_LIST[i - 1]);
				} else {
					regeneration.addPrerequisite(Names.NANOBOTS);
				}

				LIMB_OVERCLOCKING[i] = limbs.build();
				NANOBOTS_REGENERATION[i] = regeneration.build();
			}

			for (int i = 0; i < 3; i++) {
				final int level = i + 1;

				NANOBOTS_ARMOR_STRENGTH[i] = new AndroidResearchBuilder()
					.setExperienceCost(20 + i * 8)
					.withIconText(new TextComponent(String.valueOf(i + 1)))
					.withIcon(ICON_ARMOR)
					.addPrerequisite(i > 0 ? Names.NANOBOTS_ARMOR_STRENGTH_LIST[i - 1] : Names.NANOBOTS_ARMOR)
					.withName(new TranslatableComponent("android_research.overdrive_that_matters.nanobots_armor_strength", i + 1))
					.withDescription(new TranslatableComponent("android_research.overdrive_that_matters.nanobots_armor_strength.description", (i + 1) * 8, (i + 1) * 6))
					.addFeatureResult(Names.NANOBOTS_ARMOR, 0, (feature) -> {
						if (((AndroidNanobotsArmor) feature).getStrength() < level)
							((AndroidNanobotsArmor) feature).setStrength(level);
					})
					.build();

				NANOBOTS_ARMOR_SPEED[i] = new AndroidResearchBuilder()
					.setExperienceCost(20 + i * 8)
					.withIconText(new TextComponent(String.valueOf(i + 1)))
					.withIcon(ICON_ARMOR)
					.addPrerequisite(i > 0 ? Names.NANOBOTS_ARMOR_SPEED_LIST[i - 1] : Names.NANOBOTS_ARMOR)
					.withName(new TranslatableComponent("android_research.overdrive_that_matters.nanobots_armor_speed", i + 1))
					.withDescription(new TranslatableComponent("android_research.overdrive_that_matters.nanobots_armor_speed.description", (i + 1) * 8, (i + 1) * 6))
					.addFeatureResult(Names.NANOBOTS_ARMOR, 0, (feature) -> {
						if (((AndroidNanobotsArmor) feature).getSpeed() < level)
							((AndroidNanobotsArmor) feature).setSpeed(level);
					})
					.build();
			}

			LIMB_OVERCLOCKING[0].setRegistryName(Names.LIMB_OVERCLOCKING_1);
			LIMB_OVERCLOCKING[1].setRegistryName(Names.LIMB_OVERCLOCKING_2);
			LIMB_OVERCLOCKING[2].setRegistryName(Names.LIMB_OVERCLOCKING_3);
			LIMB_OVERCLOCKING[3].setRegistryName(Names.LIMB_OVERCLOCKING_4);

			NANOBOTS_ARMOR_SPEED[0].setRegistryName(Names.NANOBOTS_ARMOR_SPEED_1);
			NANOBOTS_ARMOR_SPEED[1].setRegistryName(Names.NANOBOTS_ARMOR_SPEED_2);
			NANOBOTS_ARMOR_SPEED[2].setRegistryName(Names.NANOBOTS_ARMOR_SPEED_3);

			NANOBOTS_ARMOR_STRENGTH[0].setRegistryName(Names.NANOBOTS_ARMOR_STRENGTH_1);
			NANOBOTS_ARMOR_STRENGTH[1].setRegistryName(Names.NANOBOTS_ARMOR_STRENGTH_2);
			NANOBOTS_ARMOR_STRENGTH[2].setRegistryName(Names.NANOBOTS_ARMOR_STRENGTH_3);

			NANOBOTS_REGENERATION[0].setRegistryName(Names.NANOBOTS_REGENERATION_1);
			NANOBOTS_REGENERATION[1].setRegistryName(Names.NANOBOTS_REGENERATION_2);
			NANOBOTS_REGENERATION[2].setRegistryName(Names.NANOBOTS_REGENERATION_3);
			NANOBOTS_REGENERATION[3].setRegistryName(Names.NANOBOTS_REGENERATION_4);
		}

		@SubscribeEvent
		@SuppressWarnings("unused")
		public static void register(final RegistryEvent.Register<AndroidResearchType<?>> event) {
			event.getRegistry().register(AIR_BAGS);
			event.getRegistry().register(NANOBOTS);
			event.getRegistry().register(NANOBOTS_ARMOR);
			event.getRegistry().register(EXTENDED_REACH);

			event.getRegistry().register(LIMB_OVERCLOCKING[0]);
			event.getRegistry().register(LIMB_OVERCLOCKING[1]);
			event.getRegistry().register(LIMB_OVERCLOCKING[2]);
			event.getRegistry().register(LIMB_OVERCLOCKING[3]);

			event.getRegistry().register(NANOBOTS_ARMOR_SPEED[0]);
			event.getRegistry().register(NANOBOTS_ARMOR_SPEED[1]);
			event.getRegistry().register(NANOBOTS_ARMOR_SPEED[2]);
			event.getRegistry().register(NANOBOTS_ARMOR_STRENGTH[0]);
			event.getRegistry().register(NANOBOTS_ARMOR_STRENGTH[1]);
			event.getRegistry().register(NANOBOTS_ARMOR_STRENGTH[2]);

			event.getRegistry().register(NANOBOTS_REGENERATION[0]);
			event.getRegistry().register(NANOBOTS_REGENERATION[1]);
			event.getRegistry().register(NANOBOTS_REGENERATION[2]);
			event.getRegistry().register(NANOBOTS_REGENERATION[3]);
		}
	}

	public static class Menus {
		public static final MenuType<AndroidStationMenu> ANDROID_STATION = new MenuType<>(AndroidStationMenu::new);
		public static final MenuType<MenuBatteryBank> BATTERY_BANK = new MenuType<>(MenuBatteryBank::new);
		public static final MenuType<MenuMatterDecomposer> MATTER_DECOMPOSER = new MenuType<>(MenuMatterDecomposer::new);
		public static final MenuType<MenuMatterCapacitor> MATTER_CAPACITOR_BANK = new MenuType<>(MenuMatterCapacitor::new);
		public static final MenuType<MenuPatternStorage> PATTERN_STORAGE = new MenuType<>(MenuPatternStorage::new);
		public static final MenuType<MenuMatterScanner> MATTER_SCANNER = new MenuType<>(MenuMatterScanner::new);
		public static final MenuType<MenuMatterPanel> MATTER_PANEL = new MenuType<>(MenuMatterPanel::new);
		public static final MenuType<MatterReplicatorMenu> MATTER_REPLICATOR = new MenuType<>(MatterReplicatorMenu::new);
		public static final MenuType<MenuMatterBottler> MATTER_BOTTLER = new MenuType<>(MenuMatterBottler::new);
		public static final MenuType<MenuDriveViewer> DRIVE_VIEWER = new MenuType<>(MenuDriveViewer::new);
		public static final MenuType<MenuCargoCrate> CARGO_CRATE = new MenuType<>(MenuCargoCrate::new);
		public static final MenuType<MenuDriveRack> DRIVE_RACK = new MenuType<>(MenuDriveRack::new);
		public static final MenuType<MenuItemMonitor> ITEM_MONITOR = new MenuType<>(MenuItemMonitor::new);
		public static final MenuType<MenuEnergyCounter> ENERGY_COUNTER = new MenuType<>(MenuEnergyCounter::new);
		public static final MenuType<MenuChemicalGenerator> CHEMICAL_GENERATOR = new MenuType<>(MenuChemicalGenerator::new);
		public static final MenuType<MenuPlatePress> PLATE_PRESS = new MenuType<>(MenuPlatePress::new);

		static {
			ANDROID_STATION.setRegistryName(Names.ANDROID_STATION);
			BATTERY_BANK.setRegistryName(Names.BATTERY_BANK);
			MATTER_DECOMPOSER.setRegistryName(Names.MATTER_DECOMPOSER);
			MATTER_CAPACITOR_BANK.setRegistryName(Names.MATTER_CAPACITOR_BANK);
			PATTERN_STORAGE.setRegistryName(Names.PATTERN_STORAGE);
			MATTER_SCANNER.setRegistryName(Names.MATTER_SCANNER);
			MATTER_PANEL.setRegistryName(Names.MATTER_PANEL);
			MATTER_REPLICATOR.setRegistryName(Names.MATTER_REPLICATOR);
			MATTER_BOTTLER.setRegistryName(Names.MATTER_BOTTLER);
			DRIVE_VIEWER.setRegistryName(Names.DRIVE_VIEWER);
			CARGO_CRATE.setRegistryName(Names.CARGO_CRATE);
			DRIVE_RACK.setRegistryName(Names.DRIVE_RACK);
			ITEM_MONITOR.setRegistryName(Names.ITEM_MONITOR);
			ENERGY_COUNTER.setRegistryName(Names.ENERGY_COUNTER);
			CHEMICAL_GENERATOR.setRegistryName(Names.CHEMICAL_GENERATOR);
			PLATE_PRESS.setRegistryName(Names.PLATE_PRESS);
		}

		@SubscribeEvent
		@SuppressWarnings("unused")
		public static void register(final RegistryEvent.Register<MenuType<?>> event) {
			event.getRegistry().register(ANDROID_STATION);
			event.getRegistry().register(BATTERY_BANK);
			event.getRegistry().register(MATTER_DECOMPOSER);
			event.getRegistry().register(MATTER_CAPACITOR_BANK);
			event.getRegistry().register(PATTERN_STORAGE);
			event.getRegistry().register(MATTER_SCANNER);
			event.getRegistry().register(MATTER_PANEL);
			event.getRegistry().register(MATTER_REPLICATOR);
			event.getRegistry().register(MATTER_BOTTLER);
			event.getRegistry().register(DRIVE_VIEWER);
			event.getRegistry().register(CARGO_CRATE);
			event.getRegistry().register(DRIVE_RACK);
			event.getRegistry().register(ITEM_MONITOR);
			event.getRegistry().register(ENERGY_COUNTER);
			event.getRegistry().register(CHEMICAL_GENERATOR);
			event.getRegistry().register(PLATE_PRESS);

			// OverdriveThatMatters.LOGGER.info("Registered menus");
		}

		@SubscribeEvent
		@SuppressWarnings("unused")
		public static void registerScreens(final FMLClientSetupEvent event) {
			MenuScreens.register(ANDROID_STATION, ScreenAndroidStation::new);
			MenuScreens.register(BATTERY_BANK, ScreenBatteryBank::new);
			MenuScreens.register(MATTER_DECOMPOSER, ScreenMatterDecomposer::new);
			MenuScreens.register(MATTER_CAPACITOR_BANK, ScreenMatterCapacitorBank::new);
			MenuScreens.register(PATTERN_STORAGE, ScreenPatternStorage::new);
			MenuScreens.register(MATTER_SCANNER, ScreenMatterScanner::new);
			MenuScreens.register(MATTER_PANEL, MatterPanelScreen::new);
			MenuScreens.register(MATTER_REPLICATOR, ScreenMatterReplicator::new);
			MenuScreens.register(MATTER_BOTTLER, ScreenMatterBottler::new);
			MenuScreens.register(DRIVE_VIEWER, ScreenDriveViewer::new);
			MenuScreens.register(CARGO_CRATE, ScreenCargoCrate::new);
			MenuScreens.register(DRIVE_RACK, ScreenDriveRack::new);
			MenuScreens.register(ITEM_MONITOR, ScreenItemMonitor::new);
			MenuScreens.register(ENERGY_COUNTER, ScreenEnergyCounter::new);
			MenuScreens.register(CHEMICAL_GENERATOR, ScreenChemicalGenerator::new);
			MenuScreens.register(PLATE_PRESS, ScreenPlatePress::new);

			// OverdriveThatMatters.LOGGER.info("Registered screens");
		}
	}

	public static class Stats {
		@SubscribeEvent
		@SuppressWarnings("unused")
		public static void registerVanilla(final FMLCommonSetupEvent event) {
			net.minecraft.core.Registry.register(net.minecraft.core.Registry.CUSTOM_STAT, Names.DAMAGE_ABSORBED, Names.DAMAGE_ABSORBED);
			net.minecraft.core.Registry.register(net.minecraft.core.Registry.CUSTOM_STAT, Names.HEALTH_REGENERATED, Names.HEALTH_REGENERATED);
			net.minecraft.core.Registry.register(net.minecraft.core.Registry.CUSTOM_STAT, Names.POWER_CONSUMED, Names.POWER_CONSUMED);

			net.minecraft.stats.Stats.CUSTOM.get(Names.DAMAGE_ABSORBED, StatFormatter.DIVIDE_BY_TEN);
			net.minecraft.stats.Stats.CUSTOM.get(Names.HEALTH_REGENERATED, StatFormatter.DIVIDE_BY_TEN);
			net.minecraft.stats.Stats.CUSTOM.get(Names.POWER_CONSUMED, StatFormatter.DIVIDE_BY_TEN);
		}
	}

	public static class Recipes {
		public static class MatteryRecipeType<T extends Recipe<?>> implements RecipeType<T> {
			public final ResourceLocation name;

			private MatteryRecipeType(ResourceLocation name) {
				this.name = name;
			}

			private void register() {
				net.minecraft.core.Registry.register(net.minecraft.core.Registry.RECIPE_TYPE, name, this);
			}

			@Override
			public String toString() {
				return name.toString();
			}
		}

		public static final MatteryRecipeType<PlatePressRecipe> PLATE_PRESS = new MatteryRecipeType<>(Names.PLATE_PRESS);

		@SubscribeEvent
		@SuppressWarnings("unused")
		public static void register(final FMLCommonSetupEvent event) {
			PLATE_PRESS.register();
		}

		@SubscribeEvent
		@SuppressWarnings("unused")
		public static void register(final RegistryEvent.Register<RecipeSerializer<?>> event) {
			event.getRegistry().register(PlatePressRecipeFactory.INSTANCE);
		}
	}

	public static class Tags {
		public static class Items {
			public static final IOptionalNamedTag<Item> INGOT_TRITANIUM = ItemTags.createOptional(new ResourceLocation("forge", "ingots/tritanium"), Set.of(() -> Registry.Items.TRITANIUM_INGOT));
		}
	}
}
