package ru.dbotthepony.mc.otm.block.entity

import net.minecraft.core.BlockPos
import net.minecraft.nbt.CompoundTag
import net.minecraft.network.chat.TranslatableComponent
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.AbstractContainerMenu
import net.minecraft.world.item.ItemStack
import net.minecraft.world.level.block.state.BlockState
import ru.dbotthepony.mc.otm.Registry
import ru.dbotthepony.mc.otm.block.entity.worker.BlockEntityMatteryWorker
import ru.dbotthepony.mc.otm.block.entity.worker.MachineJob
import ru.dbotthepony.mc.otm.block.entity.worker.MachineJobStatus
import ru.dbotthepony.mc.otm.capability.MatteryMachineEnergyStorage
import ru.dbotthepony.mc.otm.container.MatteryContainer
import ru.dbotthepony.mc.otm.container.MatteryContainerFilter
import ru.dbotthepony.mc.otm.core.Fraction
import ru.dbotthepony.mc.otm.menu.MenuPlatePress
import ru.dbotthepony.mc.otm.set

class BlockEntityPlatePress(p_155229_: BlockPos, p_155230_: BlockState) : BlockEntityMatteryWorker(Registry.BlockEntities.PLATE_PRESS, p_155229_, p_155230_) {
	val container = MatteryContainer(this::setChangedLight, 2)

	val handler = container.handler(object : MatteryContainerFilter {
		override fun canInsert(slot: Int, stack: ItemStack): Boolean {
			return slot != SLOT_OUTPUT
		}

		override fun canExtract(slot: Int, amount: Int, stack: ItemStack): Boolean {
			return slot != SLOT_INPUT
		}
	})

	init {
		energy = MatteryMachineEnergyStorage(this::setChangedLight, MatteryMachineEnergyStorage.MachineType.WORKER)
	}

	override fun saveAdditional(nbt: CompoundTag) {
		super.saveAdditional(nbt)
		nbt["container"] = container.serializeNBT()
	}

	override fun load(nbt: CompoundTag) {
		super.load(nbt)
		container.deserializeNBT(nbt["container"])
	}

	override fun getDefaultDisplayName() = NAME

	override fun createMenu(containerID: Int, inventory: Inventory, ply: Player): AbstractContainerMenu {
		return MenuPlatePress(containerID, inventory, this)
	}

	override fun onJobFinish(job: MachineJob): MachineJobStatus {
		val resultTag = job.data["result"] as? CompoundTag ?: return MachineJobStatus()
		val result = ItemStack.of(resultTag)

		if (result.isEmpty)
			return MachineJobStatus()

		if (!container.fullyAddItem(result, start = SLOT_OUTPUT, end = SLOT_OUTPUT)) {
			return MachineJobStatus(false, 20)
		}

		return MachineJobStatus()
	}

	override fun computeNextJob(): MachineJob? {
		val level = level ?: return null
		val recipe = level.recipeManager.getRecipeFor(Registry.Recipes.PLATE_PRESS, container, level).orElse(null) ?: return null

		val copy = container[SLOT_INPUT].copy()
		container[SLOT_INPUT].shrink(1)
		container.setChanged(SLOT_INPUT)
		copy.count = 1

		return MachineJob(copy, recipe.workTime.toDouble(), BASELINE_CONSUMPTION, CompoundTag().also {
			it["result"] = recipe.resultItem.serializeNBT()
		})
	}

	companion object {
		private val BASELINE_CONSUMPTION = Fraction(15)
		private val NAME = TranslatableComponent("block.overdrive_that_matters.plate_press")
		const val SLOT_INPUT = 0
		const val SLOT_OUTPUT = 1
	}
}
