package ru.dbotthepony.mc.otm.client.render

import com.mojang.blaze3d.vertex.PoseStack
import com.mojang.blaze3d.vertex.VertexConsumer
import com.mojang.math.Matrix4f
import com.mojang.math.Vector3f
import net.minecraft.client.gui.Font
import net.minecraft.network.chat.Component
import net.minecraft.util.FormattedCharSequence
import ru.dbotthepony.mc.otm.core.*

fun VertexConsumer.normal(vector: Vector) = normal(vector.x.toFloat(), vector.y.toFloat(), vector.z.toFloat())
fun VertexConsumer.vertex(matrix4f: Matrix4f, vector: Vector) = vertex(matrix4f, vector.x.toFloat(), vector.y.toFloat(), vector.z.toFloat())
fun VertexConsumer.color(color: RGBAColor) = color(color.r, color.g, color.b, color.a)

fun PoseStack.translate(vector: Vector) = translate(vector.x, vector.y, vector.z)
fun PoseStack.translate(vector: Vector3f) = last().pose().multiplyWithTranslation(vector.x(), vector.y(), vector.z())

fun PoseStack.rotateAroundPoint(point: Vector, axis: Vector, rotation: Float, isDegrees: Boolean = false) {
	val last = last()
	last.pose().rotateAroundPoint(point, axis, rotation, isDegrees)
	last.normal().mul(axis.rotateAroundThis(rotation, isDegrees))
}

fun PoseStack.rotateAroundPoint(point: Vector, rotation: IAngle) {
	val last = last()
	last.pose().rotateAroundPoint(point, rotation)
	// last.normal().mul(rotation.forward().rotateAroundThis(rotation))
}

fun PoseStack.rotateAroundPoint(point: Vector3f, axis: Vector, rotation: Float, isDegrees: Boolean = false) {
	val last = last()
	last.pose().rotateAroundPoint(point, axis, rotation, isDegrees)
	last.normal().mul(axis.rotateAroundThis(rotation, isDegrees))
}

fun PoseStack.rotateAroundPoint(point: Vector3f, rotation: IAngle) {
	val last = last()
	last.pose().rotateAroundPoint(point, rotation)
	// last.normal().mul(rotation.forward().rotateAroundThis(rotation))
}

fun PoseStack.translation(): Vector3f {
	return last().pose().getTranslation()
}

enum class TextAlign {
	TOP_LEFT,
	TOP_CENTER,
	TOP_RIGHT,

	CENTER_LEFT,
	CENTER_CENTER,
	CENTER_RIGHT,

	BOTTOM_LEFT,
	BOTTOM_CENTER,
	BOTTOM_RIGHT,
}

fun Font.drawAligned(poseStack: PoseStack, text: String, align: TextAlign, x: Float, y: Float, color: Int): Int {
	return when (align) {
		TextAlign.TOP_LEFT -> draw(poseStack, text, x, y, color)
		TextAlign.TOP_CENTER -> draw(poseStack, text, x - width(text) / 2f, y, color)
		TextAlign.TOP_RIGHT -> draw(poseStack, text, x - width(text), y, color)

		TextAlign.CENTER_LEFT -> draw(poseStack, text, x, y - lineHeight / 2f, color)
		TextAlign.CENTER_CENTER -> draw(poseStack, text, x - width(text) / 2f, y - lineHeight / 2f, color)
		TextAlign.CENTER_RIGHT -> draw(poseStack, text, x - width(text), y - lineHeight / 2f, color)

		TextAlign.BOTTOM_LEFT -> draw(poseStack, text, x, y - lineHeight, color)
		TextAlign.BOTTOM_CENTER -> draw(poseStack, text, x - width(text) / 2f, y - lineHeight, color)
		TextAlign.BOTTOM_RIGHT -> draw(poseStack, text, x - width(text), y - lineHeight, color)
	}
}

fun Font.drawAligned(poseStack: PoseStack, text: Component, align: TextAlign, x: Float, y: Float, color: Int): Int {
	return when (align) {
		TextAlign.TOP_LEFT -> draw(poseStack, text, x, y, color)
		TextAlign.TOP_CENTER -> draw(poseStack, text, x - width(text) / 2f, y, color)
		TextAlign.TOP_RIGHT -> draw(poseStack, text, x - width(text), y, color)

		TextAlign.CENTER_LEFT -> draw(poseStack, text, x, y - lineHeight / 2f, color)
		TextAlign.CENTER_CENTER -> draw(poseStack, text, x - width(text) / 2f, y - lineHeight / 2f, color)
		TextAlign.CENTER_RIGHT -> draw(poseStack, text, x - width(text), y - lineHeight / 2f, color)

		TextAlign.BOTTOM_LEFT -> draw(poseStack, text, x, y - lineHeight, color)
		TextAlign.BOTTOM_CENTER -> draw(poseStack, text, x - width(text) / 2f, y - lineHeight, color)
		TextAlign.BOTTOM_RIGHT -> draw(poseStack, text, x - width(text), y - lineHeight, color)
	}
}

fun Font.drawAligned(poseStack: PoseStack, text: FormattedCharSequence, align: TextAlign, x: Float, y: Float, color: Int): Int {
	return when (align) {
		TextAlign.TOP_LEFT -> draw(poseStack, text, x, y, color)
		TextAlign.TOP_CENTER -> draw(poseStack, text, x - width(text) / 2f, y, color)
		TextAlign.TOP_RIGHT -> draw(poseStack, text, x - width(text), y, color)

		TextAlign.CENTER_LEFT -> draw(poseStack, text, x, y - lineHeight / 2f, color)
		TextAlign.CENTER_CENTER -> draw(poseStack, text, x - width(text) / 2f, y - lineHeight / 2f, color)
		TextAlign.CENTER_RIGHT -> draw(poseStack, text, x - width(text), y - lineHeight / 2f, color)

		TextAlign.BOTTOM_LEFT -> draw(poseStack, text, x, y - lineHeight, color)
		TextAlign.BOTTOM_CENTER -> draw(poseStack, text, x - width(text) / 2f, y - lineHeight, color)
		TextAlign.BOTTOM_RIGHT -> draw(poseStack, text, x - width(text), y - lineHeight, color)
	}
}

fun Font.drawAligned(poseStack: PoseStack, text: String, align: TextAlign, x: Float, y: Float, color: RGBAColor): Int {
	return drawAligned(poseStack, text, align, x, y, color.toInt())
}

fun Font.drawAligned(poseStack: PoseStack, text: Component, align: TextAlign, x: Float, y: Float, color: RGBAColor): Int {
	return drawAligned(poseStack, text, align, x, y, color.toInt())
}

fun Font.drawAligned(poseStack: PoseStack, text: FormattedCharSequence, align: TextAlign, x: Float, y: Float, color: RGBAColor): Int {
	return drawAligned(poseStack, text, align, x, y, color.toInt())
}
